require 'securerandom'
require 'active_support/core_ext/hash/keys'
require 'unirest'

module Stackd
  class AuthRequest
    include Concerns::TattrAccessor
    include Concerns::RequireAttr

    tattr_accessor client: Client,
                   state: String,
                   redirect_uri: String

    class << self
      def password client, params
        if params[:username].nil?
          raise ArgumentError.new "username is required"
        end

        if params[:password].nil?
          raise ArgumentError.new "password is required"
        end

        request_token client, grant_type: 'password',
                              username: params[:username],
                              password: params[:password],
                              scope: params[:scope]
      end

      def client client, params = {}
        request_token client, grant_type: 'client_credentials',
                              scope: params[:scope]
      end

      def refresh client, token, params = {}
        if token.nil?
          raise ArgumentError.new "Token is required"
        end

        if token.refresh_token.nil?
          raise ArgumentError.new "Token must have refresh_token"
        end

        request_token client, grant_type: 'refresh_token',
                              refresh_token: token.refresh_token,
                              scope: params[:scope]
      end

      private

      def request_token client, params = {}
        if params[:grant_type].nil?
          raise ArgumentError.new "grant_type not provided"
        end

        res = Unirest.post "#{Stackd.config.api_url}/token", {
          auth: {
            user: client.id,
            password: client.secret
          },
          headers: {'Content-Type' => 'application/x-www-form-urlencoded'},
          parameters: params.reject { |k,v| v.nil? }
        }

        if res.code == 200
          token = client.tokens.new res.body.symbolize_keys

          on_grant_callbacks = \
            Stackd::Token.instance_variable_get :@_on_grant_callbacks
          on_grant_callbacks.each {|c| c.(token) }

          token
        else
          raise Error.new res.body.symbolize_keys
        end
      end
    end

    def initialize client, attrs = {}
      if client.nil?
        raise ArgumentError.new "client not provided"
      else
        self.client = client
      end

      attrs.each do |key, val|
        public_send :"#{key}=", val
      end

      if state.nil?
        self.state = SecureRandom.urlsafe_base64 9
      end
    end

    def url params = {}
      uri = auth_url

      uri.query_values = {
        response_type: 'code',
        client_id: client.id,
        state: state,
        redirect_uri: params[:redirect_uri],
        scope: params[:scope]
      }.reject { |k,v| v.nil? }

      uri.to_s
    end

    def callback params
      unless params[:code] || params[:error]
        raise ArgumentError.new 'no code or error'
      end

      if params[:state] != state
        raise StateMismatchError.new "#{params[:state]} != #{state}"
      end

      if params[:error]
        raise Error.new params
      else
        self.class.send :request_token, client, {
          grant_type: 'authorization_code',
          code: params[:code],
          redirect_uri: redirect_uri
        }
      end
    end

    private

    def auth_url
      Addressable::URI.parse Stackd.config.auth_url
    end

    class StateMismatchError < StandardError
    end
  end
end
