require 'json'
require 'uri'
require 'unirest'
require 'active_support/core_ext/hash/keys'

module Stackd
  module Util
    module HTTP
      def self.get credentials, url
        request_without_body :get, credentials, url
      end

      def self.post credentials, url, body, content_type
        request_with_body :post, credentials, url, body, content_type
      end

      def self.put credentials, url, body, content_type
        request_with_body :put, credentials, url, body, content_type
      end

      def self.patch credentials, url, body, content_type
        request_with_body :patch, credentials, url, body, content_type
      end

      def self.delete credentials, url
        request_without_body :delete, credentials, url
      end

      private

      def self.request_without_body method, credentials, url
        validate_params! credentials, url

        basic_auth = basic_auth_for credentials
        headers = headers_for credentials

        res = Unirest.public_send method, url, headers: headers,
                                               auth: basic_auth

        process_response res
      end

      def self.request_with_body method, credentials, url, body, content_type
        validate_params! credentials, url, body, content_type

        basic_auth = basic_auth_for credentials
        headers = headers_for credentials, content_type
        parameters = parameters_for body, content_type

        res = Unirest.public_send method, url, headers: headers,
                                               auth: basic_auth,
                                               parameters: parameters

        process_response res
      end

      def self.validate_params! credentials, url, body = nil, content_type = nil
        unless creds_are_valid? credentials
          raise ArgumentError.new \
            "credentials must be a Stackd::Token, Hash with :user + " + \
              ":password, credentials: #{credentials.inspect}"
        end

        unless url.is_a? String
          raise ArgumentError.new "url must be String. url: #{url.inspect}"
        end

        unless body.nil? || body.is_a?(Hash)
          raise ArgumentError.new \
            "body must be a Hash or nil. body: #{body.inspect}"
        end

        unless content_type.nil? || [:json, :form].include?(content_type)
          raise ArgumentError.new \
            "content_type must be :json, :form, or nil. " + \
              "content_type: #{content_type.inspect}"
        end
      end

      def self.creds_are_valid? credentials
        credentials.nil? || \
        credentials.is_a?(Client) || \
        credentials.is_a?(Token) || \
        (credentials.is_a?(Hash) && \
         credentials[:user].is_a?(String) && \
         credentials[:password].is_a?(String))
      end

      def self.basic_auth_for credentials
        if credentials.is_a? Hash
          credentials
        elsif credentials.is_a? Client
          { user: credentials.id, password: credentials.secret }
        end
      end

      def self.headers_for credentials, content_type = nil
        headers = {}

        if credentials.is_a? Token
          headers.merge!(
            { 'Authorization' => "Bearer #{credentials.access_token}" })
        end

        unless content_type.nil?
          content_type_value = content_type == :json \
            ? 'application/json' \
            : 'application/x-www-form-urlencoded'

          headers.merge!({ 'Content-Type' => content_type_value })
        end

        headers
      end

      def self.parameters_for body, content_type
        if content_type == :json
          JSON.generate body
        else
          body
        end
      end

      def self.process_response res
        if res.code >= 500
          raise ServerError.new res.body.deep_symbolize_keys
        elsif res.code >= 400
          raise Error.new res.body.deep_symbolize_keys
        else
          res.body.deep_symbolize_keys
        end
      end
    end
  end
end
